@@ -98,6 +98,9 @@ end
 
 
 function DocView:get_scrollable_size()
+  if not config.scroll_past_end then
+    return self:get_line_height() * (#self.doc.lines) + style.padding.y * 2
+  end
   return self:get_line_height() * (#self.doc.lines - 1) + self.size.y
 end
 
@@ -408,10 +411,12 @@ function DocView:draw_overlay()
     local T = config.blink_period
     for _, line, col in self.doc:get_selections() do
       if line >= minline and line <= maxline
-      and (core.blink_timer - core.blink_start) % T < T / 2
       and system.window_has_focus() then
-        local x, y = self:get_line_screen_position(line)
-        self:draw_caret(x + self:get_col_x_offset(line, col), y)
+        if config.disable_blink
+        or (core.blink_timer - core.blink_start) % T < T / 2 then
+          local x, y = self:get_line_screen_position(line)
+          self:draw_caret(x + self:get_col_x_offset(line, col), y)
+        end
       end
     end
   end
@@ -434,7 +439,9 @@ function DocView:draw()
 
   local pos = self.position
   x, y = self:get_line_screen_position(minline)
-  core.push_clip_rect(pos.x + gw, pos.y, self.size.x, self.size.y)
+  -- the clip below ensure we don't write on the gutter region. On the
+  -- right side it is redundant with the Node's clip.
+  core.push_clip_rect(pos.x + gw, pos.y, self.size.x - gw, self.size.y)
   for i = minline, maxline do
     self:draw_line_body(i, x, y)
     y = y + lh
